import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.filter.PlugInFilter;

import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.Checkbox;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.System;
import java.lang.Math.*;

public class CoxGroupHawk implements PlugInFilter {
	
	ImagePlus window;
	String arg;

    public int setup(String arg_, ImagePlus img) {
		window = img;
		arg=arg_;

        return STACK_REQUIRED +NO_CHANGES + DOES_16 + DOES_32 + DOES_8G;
    }

	public void run(ImageProcessor ip) {

		//The non-config file parameters are the range of frames to  operate on
		//and the pixel size in nm (the config works in terms of FWHM in pixels).
		//These have to be sepficied whether the basic or advanced dialog is used.
		int  firstfr;
		int lastfr;
		double pixel_size_in_nm;

		ImageStack s = window.getStack();

		HawkDialog gd = new HawkDialog();

		gd.showDialog();

		if(!gd.wasOKed())
			return;

		int levels = (int)gd.getNextNumber();
		String neg = gd.getNextChoice();
		
		String ordering = gd.getNextChoice();

		System.out.println("Levels = " + levels);
		System.out.println("Images = " + s.getSize());
		System.out.println("Neg = " + neg);
		System.out.println("Ordering = " + ordering);
		
		int negi=0;
		if(neg.equals("ABS"))
			negi = 1;
		else if(neg.equals("Separate"))
			negi=2;

		boolean interleave = false;
		if(ordering.equals("Group temporally"))
			interleave = true;

		System.out.println("Neg = " + negi);
		System.out.println("Interleave = " + interleave);
		ImageStack filtered = new HawkProcessor().filter(s, levels, negi, interleave);

		ImagePlus view = new ImagePlus("HAWK processed " + window.getTitle(), filtered);
		view.show();
		System.out.println("Done");
	}

	static 
	{
		new CoxGroupCheckForUpdates().run("boot");
	}
};


class HawkProcessor
{
	
	public ImageStack filter(ImageStack in, int levels, final int neg, final boolean interleave)
	{
		//The kernel is -1,1 times 1<<level
		//In other words, kernel width W = s<<level
		//levels count from 0

		//The stack size is N. So, N-W+1 kernels fit within it.

		final int N =in.getSize();
		final int w =in.getWidth();
		final int h =in.getHeight();

		ImageStack out = ImageStack.create(w, h, 0, 32);
		
		//size is always 1 or more...
		out.deleteLastSlice();
		
		//Compute the number of frames in the output
		int target_size=0;
		for(int l=0; l < levels; l++)
		{
			final int kernel_w = 2 << l;
			target_size += (N-kernel_w+1);
		}
		if(neg == 2)
			target_size+=2;
		
		//The options are to either have all frames with kernel 1, then with kernel 2 
		//and so on, or to order the output so that all results with the same central
		//frame go together. This may allow drift correction to be used.
		//
		//The lazy simple way to do it is to do the interleaving at the end
		ArrayList<ArrayList<FloatProcessor>> results = new ArrayList<ArrayList<FloatProcessor>>();

		if(interleave)
			for(int i=0; i < N; i++)
				results.add(new ArrayList<FloatProcessor>());
		else
			results.add(new ArrayList<FloatProcessor>());

		for(int l=0; l < levels; l++)
		{
			final int kernel_w = 2 << l;
			final int kernel_half_w = 1 << l;
			final float normalization_factor = (float)Math.sqrt(kernel_w);

			for(int s=0; s < N-kernel_w+ 1; s++)
			{
				FloatProcessor fp = new FloatProcessor(w, h);
				FloatProcessor fn = new FloatProcessor(w, h);

				for(int r=0; r < h; r++)
					for(int c=0; c<w; c++)
					{
						float sum=0;
						for(int i=0; i < kernel_half_w; i++)
							sum += in.getVoxel(c, r, s+i) - in.getVoxel(c, r, s+i+kernel_half_w);
						
						if(neg ==1)
						{
							fp.setf(c, r, (float)Math.abs(sum));
						}
						else if(neg == 2)
						{
							if(sum > 0)
								fp.setf(c, r, sum);
							else
								fn.setf(c, r, -sum);
						}
					}
				
				//Increment the top left pixel by 1. this makes sure that the image
				//has some variation so that analysis algorithms won't conk out when 
				//fed with simulated data that happens to give an image of all zeros.
				fp.setf(0,0,fp.getf(0,0)+1);
				fn.setf(0,0,fn.getf(0,0)+1);

				int pos_to_add = 0;

				if(interleave)
					pos_to_add = s + kernel_half_w - 1; // This is the "central" frame
				else
					pos_to_add = 0; //Otherwise add things in the order they're computed

				results.get(pos_to_add).add(fp);
				
				//out.addSlice(fp);

				if(neg == 2)
				{
					results.get(pos_to_add).add(fn);
					//out.addSlice(fn);
				}

				IJ.showProgress(out.getSize(), target_size);
			}
		}

		
		for(int i=0; i < results.size(); i++)
			for(int j=0; j < results.get(i).size(); j++)
				out.addSlice(results.get(i).get(j));

		IJ.showProgress(0,0);

		return out;
	}
}


///Dialog box for starting HAWK
///The dialog highlights bad things in red.
///@ingroup gPlugin
//Not pretty. Should I use dialogItemChanged rather than textValueChanged???
class HawkDialog extends GenericDialog
{
	Color levels_c, bg;

	HawkDialog()
	{
		super("HAWK");

		addNumericField("Number of Levels", 3, 0);
		addChoice("Negative values", new String[]{"ABS", "Separate"}, "Separate");
		addChoice("Output order", new String[]{"Group by level", "Group temporally"}, "Kernel size");

		addTextAreas("",null, 8,30);

		getTextArea1().setEditable(false);
		levels_c = getLevelsField().getBackground(); //Get the default background colour
		bg = getBackground(); //Dialog background color
		getTextArea1().removeTextListener(this); //To prevent event thrashing when we write messages

		//Process initial warnings
		textValueChanged(null);
	}

	int parseInt(String s)
	{
		try
		{
			return Integer.parseInt(s);
		}
		catch(Exception e)
		{
			return 0;
		}
	}
	

	TextField getLevelsField()
	{
		return (TextField)(getNumericFields().get(0));
	}

	
	double get(TextField f)
	{	
		return parseDouble(f.getText());
	}
	
	double getLevels()
	{
		return parseDouble(getLevelsField().getText());
	}
	
	public void itemStateChanged(ItemEvent e)
	{	
		change();
	}

	public void textValueChanged(TextEvent e)
	{
		change();
	}
	void change()
	{
		boolean bad=false;
		String err = "";
		
		//Make the levels a positive iteger
		int levels= (int)Math.max(1.0, getLevels());

		if(levels < 1 && !getLevelsField().getText().equals(""))
			getLevelsField().setText(Integer.toString( (int)Math.max(1.0, levels)));

		//               012345678901234567890123456789012345678901234567890
		if(getLevels() > 5)
		{
			bad=true;
			err = err + "Warning: Diminishing returns and slow runs about 5 levels";
			getLevelsField().setBackground(Color.RED);
		}
		else
		{
			getLevelsField().setBackground(levels_c);
		}

		if(!err.equals(""))
			getTextArea1().setBackground(Color.RED);
		else
			getTextArea1().setBackground(bg);


		getTextArea1().setText(err);
		repaint();

	}
}


