import ij.*;
import ij.process.*;
import ij.gui.*;
import java.awt.*;
import ij.plugin.*;
import ij.plugin.frame.*;
import ij.plugin.filter.PlugInFilter;
import ij.*;
import ij.io.*;
import ij.plugin.*;
import ij.plugin.filter.*;

import java.awt.image.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import java.net.*;
import java.io.*;
import javax.swing.*;
import javax.swing.event.*;
import java.lang.InterruptedException;
import java.lang.System;
import java.lang.Math.*;
import java.lang.reflect.*;

//Please accept my apologies for poor Java code.
//This is my first ever Java program.



//The orginal control panel from 3B has some stuff baked in for running 3B obviously.
//
//Uderneath it's a quite simple concept with a run method, a stop method, a way
//of registering callbacks and some status stuff. Anyway we're keeping it here
//as an interface (unused) so we don't have to rip it out and so that we can merge
//3B back into the Cox group plugin later.
abstract class ThreeBRunnerInterface implements Runnable
{
	abstract void register(CoxGroupReconstructionViewer panel);
	abstract boolean has_stopped();
	abstract void stop_thread();
};


class ThreeBGlobalConstants
{
	public static int critical_iterations=200;
};


///Utility class to hold a pair of strings
///
///@ingroup gPlugin
class SPair
{
	public String a, b;
}

///Utility calss to hold a number of handy static functions.
///@ingroup gPlugin

///Basic spot class, simply contains coordinates.
///@ingroup gPlugin
class Spot
{
	double x, y, sigma;
	Spot()
	{
		sigma = 1;
	}

	Spot(double xx, double yy)
	{
		x=xx;
		y=yy;
		sigma = 1;
	}

	Spot(double xx, double yy, double ss)
	{
		x=xx;
		y=yy;
		sigma = ss;
	}
}


class Reconstruction
{
	///Compute a brand new reconstruction.
	static FloatProcessor reconstruct(Rectangle roi, double zoom, double reconstruction_blur_fwhm, double pixel_size_in_nm, ArrayList<Spot> pts)
	{
		return reconstruct(roi, zoom, reconstruction_blur_fwhm, pixel_size_in_nm, pts, -1e9);
	}

	static FloatProcessor reconstruct(Rectangle roi, double zoom, double reconstruction_blur_fwhm, double pixel_size_in_nm, ArrayList<Spot> pts, double sigma_filter_nm)
	{
		//Reconstruct an image which is based around the ROI in size
		int xoff = (int)roi.x;
		int yoff = (int)roi.y;

		//Apparently combining a 1x1 rectangle at 0,0 and 24,24 gives a total width of... 24.
		//lol.
		int xsize = (int)Math.round((roi.width + 1) * zoom);
		int ysize = (int)Math.round((roi.height + 1) * zoom);
		
		//New blank image set to zero
		FloatProcessor reconstructed = new FloatProcessor(xsize, ysize);

		final double sigma_filter = sigma_filter_nm / pixel_size_in_nm;

		for(int i=0; i < pts.size(); i++)
		{
			
			//Increment the count 
			if(pts.get(i).sigma >= sigma_filter)
			{
				int xc = (int)Math.floor((pts.get(i).x-xoff) * zoom + 0.5);
				int yc = (int)Math.floor((pts.get(i).y-yoff) * zoom + 0.5);
				float p = reconstructed.getPixelValue(xc, yc);
				reconstructed.putPixelValue(xc, yc, p+1);
			}
		}

		double blur_sigma = reconstruction_blur_fwhm / (2 * Math.sqrt(2 * Math.log(2))) * zoom / pixel_size_in_nm;

		(new GaussianBlur()).blurGaussian(reconstructed, blur_sigma, blur_sigma, 0.005);

/*
		//Make an image

		int size =xsize;
		
		//New image
		FloatProcessor reconstructed = new FloatProcessor(size, size);
		for(int y=0; y < size; y++)
			for(int x=0; x < size; x++)
				if(x % 20 < 10 != y % 20 < 10)
					reconstructed.putPixelValue(x, y, x * 255.0 / size);

*/
		return reconstructed;
	}
}


///Control panel for running 3B plugin and providing interactive update.
///@ingroup gPlugin
class CoxGroupReconstructionViewer extends JFrame implements WindowListener
{

	///Listener class which triggers a complete redraw. Since all redraws are
	///pretty much equal, there is no need to distinguish them.
	///@ingroup gPlugin
	class CoxGroupReconstructionViewerSomethingChanges implements ChangeListener
	{
		private CoxGroupReconstructionViewer c;
		public CoxGroupReconstructionViewerSomethingChanges(CoxGroupReconstructionViewer c_)
		{
			c = c_;
		}
		
		
		public void stateChanged(ChangeEvent e)
		{
			c.send_update_canvas_event();
		}
	}

	///This class makes a floating point slider bar with an edit box next to 
	///it for more precision. Also has a reciprocal option for inverse, reciprocal scaling./
	///@ingroup gPlugin
	//That's not at all bodged in in an unpleasant way.
	class FloatSliderWithBox extends JPanel {

		private JSlider slider;
		private JTextField number;
		private JLabel label;
		private GridBagConstraints completePanelConstraints_;

		private int steps=1000000;
		private double min, max;
		private String text;
		private String units;
		private String format = "%8.3f";

		private double value;
		private boolean reciprocal;

		public FloatSliderWithBox(String text_, double min_, double max_, double value_, int cols, boolean rec_)
		{
			
			super( new GridBagLayout() );

			reciprocal = rec_;
			min=min_;
			max=max_;
			text=text_;
			value = value_;

			if(reciprocal)
			{
				min=1/max_;
				max=1/min_;
			}

			slider = new JSlider(0, steps);

			label = new JLabel();
			
			number = new JTextField(cols);

			//Assemble into a panel
			completePanelConstraints_ = new GridBagConstraints();
			completePanelConstraints_.fill = GridBagConstraints.HORIZONTAL;

			completePanelConstraints_.gridx = 0;
			completePanelConstraints_.gridy = 0;
			completePanelConstraints_.weightx=1;
			this.add( slider, completePanelConstraints_ );

			completePanelConstraints_.gridx = 2;
			completePanelConstraints_.gridy = 0;
			completePanelConstraints_.weightx=0;
			this.add(label, completePanelConstraints_ );

			completePanelConstraints_.gridx = 1;
			completePanelConstraints_.gridy = 0;
			completePanelConstraints_.weightx=0;
			//Add some space to the left to move away from slider slightly
			//And some more to the bottom to help with the way they are displayed
			completePanelConstraints_.insets = new Insets(0,5,10,0);
			this.add( number, completePanelConstraints_ );

			this.setBorder(BorderFactory.createTitledBorder(text));


			slider.addChangeListener(new SliderChanged(this));
			number.addActionListener(new TextChanged(this));
			setValue(value);

		}

		public FloatSliderWithBox setUnits(String s)
		{
			units = s;
			setValue(value);
			return this;
		}

		public FloatSliderWithBox setFormat(String s)
		{
			format = s;
			setValue(value);
			return this;
		}

		public void addChangeListener(ChangeListener changeListener){
			slider.addChangeListener( changeListener );
			return;
		}

		void setValue(double v)
		{
			value = v;
			if(reciprocal)
				slider.setValue((int)Math.round(steps * (1/value-min)/(max-min)));
			else
				slider.setValue((int)Math.round(steps * (value-min)/(max-min)));
				
			number.setText(String.format(format, value));
			label.setText(units);
		}

		double getValue()
		{
			return value;
		}

		public double get_value_from_slider() 
		{ 
			if(reciprocal)
				return 1/((slider.getValue() * 1.0 / steps) * (max - min) + min);
			else
				return (slider.getValue() * 1.0 / steps) * (max - min) + min;
		}
		public double get_value_from_text() 
		{ 
			return Double.parseDouble(number.getText());
		}

		class SliderChanged implements ChangeListener
		{
			FloatSliderWithBox f;
			SliderChanged(FloatSliderWithBox f_)
			{
				f = f_;
			}

			public void stateChanged(ChangeEvent e)
			{
				f.setValue(f.get_value_from_slider());
			}
		}

		class TextChanged implements ActionListener
		{
			FloatSliderWithBox f;
			TextChanged(FloatSliderWithBox f_)
			{
				f = f_;
			}

			public void actionPerformed(ActionEvent e)
			{
				f.setValue(f.get_value_from_text());
			}
		}

	}


	///Close button issues a window close event. Actual closing logic is then
	///done in the close event handler.
	///@ingroup gPlugin
	class CloseButtonListener implements ActionListener
	{
		private JFrame f;
		public CloseButtonListener(JFrame fr)
		{
			f = fr;
		}
		
		
		public void actionPerformed(ActionEvent e)
		{
			//Voodoo from Stack Overflow
			WindowEvent wev = new WindowEvent(f, WindowEvent.WINDOW_CLOSING);
			Toolkit.getDefaultToolkit().getSystemEventQueue().postEvent(wev);
		}
	}

	///Stop 3B thread
	///@ingroup gPlugin
	class StopButtonListener implements ActionListener
	{
		private CoxGroupReconstructionViewer t;
		public StopButtonListener(CoxGroupReconstructionViewer t_)
		{
			t = t_;
		}

		public void actionPerformed(ActionEvent e)
		{
			t.issue_stop();
		}
	}

	///Listener for the export button.
	///@ingroup gPlugin
	class ExportButtonListener implements ActionListener
	{
		private CoxGroupReconstructionViewer c;
		public ExportButtonListener(CoxGroupReconstructionViewer c_)
		{
			c = c_;
		}

		public void actionPerformed(ActionEvent e)
		{
			c.export_reconstruction_as_ij();
		}
	}
	private ImagePlus   linear_reconstruction;  //Reconstructed image 
	private ImageCanvas canvas;
	private JButton stopButton, exportButton, closeButton;
	private Color exportButtonColor;
	private JLabel status, time_msg;
	private FloatSliderWithBox blur_fwhm;
	private FloatSliderWithBox reconstructed_pixel_size;
	private FloatSliderWithBox sigma_filter;

	private JPanel complete;
	private JPanel buttons;
	private JScrollPane scroll;

	private ThreeBRunnerInterface tbr;
	private ArrayList<Spot> pts;

	private JLabel image_label;
	private ImageIcon icon;


	private Rectangle roi;
	private double zoom=0.01; //Sets initial zoom small,  so the ImageCanvas will always be bigger than its initial size
	                          //otherwise it always puts up the zooming rectangle :(
	private double reconstruction_blur_fwhm=.5;
	private double pixel_size_in_nm;
	private double sigma_filter_nm;
	private String filename;

	//Number of iterations. Used to colourize export button and provide warnings.
	//Architecture is now getting quite messy.
	private int iterations = 0;
	
	private boolean show_sigma_filter;
	
	void set_reconstruction_blur_fwhm(double r)
	{
		blur_fwhm.setValue(r);
	}
	void set_reconstructed_pixel_size(double z)
	{
		reconstructed_pixel_size.setValue(z);
	}
	void set_sigma_filter(double z)
	{
		sigma_filter.setValue(z);
	}

	CoxGroupReconstructionViewer(Rectangle roi_, double ps_, String filename_, ThreeBRunnerInterface tbr_, boolean sig_)
	{
		//Constract superclass
		super(filename_);

		tbr = tbr_;
		filename=filename_;
		show_sigma_filter=sig_;

		roi = roi_;
		pixel_size_in_nm = ps_;
		pts = new ArrayList<Spot>();

	
		//Now generate the dialog box
		complete = new JPanel(new GridBagLayout());

		//Create the image viewer
		linear_reconstruction = new ImagePlus();
		linear_reconstruction.setProcessor(reconstruct());
	//	canvas = new ImageCanvas(linear_reconstruction);
		
		//Make the image scrollable. This seems to work, if the correct cargo-culting
		//is performed with the gridbaglayout fill constraints. I don't really understand why.


		BufferedImage im = new BufferedImage(100,100,BufferedImage.TYPE_BYTE_GRAY);
		icon = new ImageIcon(im);
		image_label = new JLabel(icon);
		

		scroll = new JScrollPane(image_label);
		//scroll.add(image_label);

		//scroll.add(canvas);

		complete.add(scroll, canvas_pos());

		//Create the status message
		status = new JLabel();
		if(tbr != null)
			set_status("Running.");
		else
			set_status("Using loaded data.");
		complete.add(status, status_pos());
		
		//Create the ETA massage
		time_msg = new JLabel();
		if(tbr != null)
			set_time("unknown");
		else
			set_time("not running");
		complete.add(time_msg, time_pos());
		

		
		//The two control sliders
		blur_fwhm = new FloatSliderWithBox("Reconstruction blur FWHM", 0, 150, 100.0, 5, false);
		blur_fwhm.setUnits("nm").setFormat("%5.1f");
		blur_fwhm.addChangeListener(new CoxGroupReconstructionViewerSomethingChanges(this));
		complete.add(blur_fwhm, blur_pos());

		reconstructed_pixel_size = new FloatSliderWithBox("Reconstructed pixel size", 2, 300, 50.0, 5, true);
		reconstructed_pixel_size.setUnits("nm").setFormat("%5.1f");
		reconstructed_pixel_size.addChangeListener(new CoxGroupReconstructionViewerSomethingChanges(this));
		complete.add(reconstructed_pixel_size, pixel_size_pos());
		

		if(show_sigma_filter)
		{
			sigma_filter = new FloatSliderWithBox("Remove small localisations", 0, 500, 100.0, 5, false);
			sigma_filter.setUnits("nm").setFormat("%5.1f");
			sigma_filter.addChangeListener(new CoxGroupReconstructionViewerSomethingChanges(this));
			complete.add(sigma_filter, sigma_filter_pos());
		}

		//The button bar
		buttons = new JPanel();
		
		exportButton = new JButton("Export...");
		exportButton.addActionListener(new ExportButtonListener(this));
		buttons.add(exportButton);
		exportButtonColor = exportButton.getBackground();
		
		//No point in having a stop button if this is just a viewer
		if(tbr != null)
		{
			stopButton = new JButton("Stop");
			stopButton.addActionListener(new StopButtonListener(this));
			buttons.add(stopButton);
		}

		closeButton = new JButton("Close");
		closeButton.addActionListener(new CloseButtonListener(this));
		buttons.add(closeButton);


		complete.add(buttons, buttons_pos());

		getContentPane().add(complete);
		
		//This class knows how to handle window close events.
		//Don't close the window by default, so we can try to bring up a 
		//confirmation dialog.
		addWindowListener(this);
		setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
		

		//Cargo culting...
		pack();
		setVisible(true);
		validate();
		setSize(600,600);
		validate();


		//Start the 3B thread, if we have one to run.
		if(tbr != null)
		{
			tbr.register(this);
			Thread t = new Thread(tbr);
			t.start();
		}
	}

	//Functions for generating constraings for gridbaglayout
	private GridBagConstraints canvas_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 0;
		g.fill = GridBagConstraints.BOTH;
		g.weightx=100;
		g.weighty=100;
		return g;
	}

	private GridBagConstraints status_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 1;
		g.anchor = GridBagConstraints.FIRST_LINE_START;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}

	private GridBagConstraints time_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 2;
		g.anchor = GridBagConstraints.FIRST_LINE_START;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}
	private GridBagConstraints buttons_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		if(show_sigma_filter)
			g.gridy = 6;
		else
			g.gridy = 5;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}

	private GridBagConstraints sigma_filter_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 5;
		g.anchor = GridBagConstraints.FIRST_LINE_START;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}

	private GridBagConstraints pixel_size_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 4;
		g.anchor = GridBagConstraints.FIRST_LINE_START;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}
	
	private GridBagConstraints blur_pos()
	{
		GridBagConstraints g = new GridBagConstraints();
		g.gridx = 0;
		g.gridy = 3;
		g.anchor = GridBagConstraints.FIRST_LINE_START;
		g.fill = GridBagConstraints.BOTH;
		return g;
	}
	
	private void set_status(String s)
	{
		status.setText("Status: " + s);
	}

	private void set_time(String s)
	{
		time_msg.setText("Estimated time: " + s);
	}
	
	///Generic redraw function which recomputes the reconstruction.
	///this is a synchronized method since it makes use of the shared writable
	///array of points.
	private synchronized void update_canvas()
	{
		reconstruction_blur_fwhm = blur_fwhm.getValue();

		
		if(show_sigma_filter)
			sigma_filter_nm = sigma_filter.getValue();
		else
			sigma_filter_nm = -1e9; //Really, any negative number should do!

		zoom = pixel_size_in_nm / reconstructed_pixel_size.getValue();

		linear_reconstruction.setProcessor(reconstruct());
		linear_reconstruction.updateImage();
		linear_reconstruction.updateAndRepaintWindow();
		//canvas.repaint();	
		//Update the canvas size and viewport to prevent funny things with zooming.
		//canvas.setDrawingSize(linear_reconstruction.getWidth(), linear_reconstruction.getHeight());
		//canvas.setSourceRect(new Rectangle(linear_reconstruction.getWidth(), linear_reconstruction.getHeight()));
		//validate();
		//pack();

		if(iterations >= ThreeBGlobalConstants.critical_iterations)
		{
			exportButton.setBackground(exportButtonColor);
		}
		else
		{
			exportButton.setBackground(Color.RED);
		}

		
		icon.setImage(linear_reconstruction.getImage());
		image_label.setIcon(icon);
		image_label.setSize(linear_reconstruction.getWidth(), linear_reconstruction.getHeight());


		//Cargo culting...
		//Without non resizing changes (i.e. blur) do not cause
		//window repaints when both scrollbars are present. 
		image_label.repaint();
		scroll.repaint();

		//Recompues scrollbars correctly.
		//Don't have pack: that causes a resize!
		validate();
	}


	//Window event methods. We have to implement all of these to 
	//be a window listener. We don't care about any of these.
	public void windowOpened(WindowEvent e) { }
	public void windowClosed(WindowEvent e){}
	public void windowDeactivated(WindowEvent e) { }
	public void windowActivated(WindowEvent e) { }
	public void windowDeiconified(WindowEvent e) { }
	public void windowIconified(WindowEvent e) { }

	///Send a stop message to the thread if the window is closed
	public void windowClosing(WindowEvent e)
	{
		if(tbr != null && !tbr.has_stopped())
		{
			int confirmed = JOptionPane.showConfirmDialog(null, "3B still running! Closing will terminate the run. Really close?", "User Confirmation", JOptionPane.YES_NO_OPTION);

			if (confirmed == JOptionPane.YES_OPTION)
			{
				dispose();
				tbr.stop_thread();
			}

		}
		else
		{
			//OK to close
			dispose();
		}
	}

	///Generate a fresh ImageJ image in a standard imageJ window, so that ti can be 
	///manipulated using the usual ImageJ commands. The image is size calibrated, so
	///scalebars are easy to add.
	void export_reconstruction_as_ij()
	{
		ImageProcessor export = linear_reconstruction.getProcessor().duplicate();
		ImagePlus export_win = new ImagePlus(filename + " reconstruction", export);
		export_win.getCalibration().pixelWidth = reconstructed_pixel_size.getValue();
		export_win.getCalibration().pixelHeight = reconstructed_pixel_size.getValue();
		export_win.getCalibration().setXUnit("nm");
		export_win.getCalibration().setYUnit("nm");
		export_win.show();
		export_win.updateAndDraw();
	}
	
	///Compute a brand new reconstruction.
	private synchronized FloatProcessor reconstruct()
	{
	/*
		//Reconstruct an image which is based around the ROI in size
		int xoff = (int)roi.x;
		int yoff = (int)roi.y;

		//Apparently combining a 1x1 rectangle at 0,0 and 24,24 gives a total width of... 24.
		//lol.
		int xsize = (int)Math.round((roi.width + 1) * zoom);
		int ysize = (int)Math.round((roi.height + 1) * zoom);
		
		//New blank image set to zero
		FloatProcessor reconstructed = new FloatProcessor(xsize, ysize);

		for(int i=0; i < pts.size(); i++)
		{
			
			//Increment the count 
			int xc = (int)Math.floor((pts.get(i).x-xoff) * zoom + 0.5);
			int yc = (int)Math.floor((pts.get(i).y-yoff) * zoom + 0.5);
			float p = reconstructed.getPixelValue(xc, yc);
			reconstructed.putPixelValue(xc, yc, p+1);
		}

		double blur_sigma = reconstruction_blur_fwhm / (2 * Math.sqrt(2 * Math.log(2))) * zoom / pixel_size_in_nm;

		(new GaussianBlur()).blurGaussian(reconstructed, blur_sigma, blur_sigma, 0.005);
		return reconstructed;
		*/

		return Reconstruction.reconstruct(roi, zoom, reconstruction_blur_fwhm, pixel_size_in_nm, pts, sigma_filter_nm);
	}


	public void issue_stop()
	{
		if(tbr != null)
			tbr.stop_thread();
		stopButton.setEnabled(false);
	}

	//Below here are callbacks used by the ThreeBRunnerInterface thread.

	///Callback issued by the runnre thread which appends the latest set of 
	///points to the array. This is synchronized since it involves the shared
	///writable array of points.
	synchronized public void append(final ArrayList<Spot> p, int its)
	{
		for(int i=0; i < p.size(); i++)
			pts.add(p.get(i));

		iterations = its;
	}
	
	
	///Callback for causing a recomputation of the reconstruction. 
	///This is a safe method for telling the class to recompute and redraw the reconstruction.
	///It can be called from anywhere, but ensures that the redraw happens in the GUI thread.
	public void send_update_canvas_event()
	{
		SwingUtilities.invokeLater(
			new Runnable() {
				final public void run() {
						update_canvas();
					}
				}
		);
	}

	
	///Callback which causes a fatal termination of the 3B control panel. This is generally caused
	///by something changing such that the error condition was not seen in the Java code, but was seen
	///in C++. Ought never to be called.
	public void die(final String s)
	{
		SwingUtilities.invokeLater(
			new Runnable() {
				final public void run() {
						ij.IJ.showStatus("3B run terminated due to an error.\n");
						JOptionPane.showMessageDialog(null, "Error: "+s, "Fatal error", JOptionPane.ERROR_MESSAGE);
						dispose();
					}
				}
		);
	}
	
	///Callback to update the status message safely.
	public void send_status_text_message(final String s)
	{
		SwingUtilities.invokeLater(
			new Runnable(){
				final public void run() {
						set_status(s);
					}
				}
		);
	}


	///Callback to update the status message safely.
	public void send_time_text_message(final String s)
	{
		SwingUtilities.invokeLater(
			new Runnable(){
				final public void run() {
						set_time(s);
					}
				}
		);
	}



};
